"""This module contains the QueryError class."""

import sys

import click


class QueryError(Exception):
    """Exception raised for errors during query data processing."""

    def __init__(self, resource: str, detail: str) -> None:
        """Initialize the QueryError class."""
        self.resource = resource
        self.detail = detail
        self.message = f"QueryError: {resource} - {detail}"
        super().__init__(self.message)

    def handler(self) -> None:
        """Handle the query error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the CacheIOError class."""

import sys
from enum import Enum

import click


class CacheIOErrorMessage(Enum):
    """Enum class for cacheIO error types."""

    VERSION_FILE_READ_FAILED = "Failed to read version cache file. Please try again later."
    MATCH_SELECTION_FILE_READ_FAILED = "Failed to read match selection cache file. Please try again later."
    WRITE_FAILED = "Failed to write cache file. Please try again later."
    MATCH_NOT_SELECTED = "Please select a match first."


class CacheIOError(Exception):
    """Exception raised for errors in cache file I/O."""

    def __init__(self, error_type: CacheIOErrorMessage) -> None:
        """Initialize the CacheIOError class."""
        super().__init__(error_type.value)

    def handler(self) -> None:
        """Handle the cache file I/O error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the functions related to help command."""

import click

from opthub_client.controllers.download import download
from opthub_client.controllers.login import login
from opthub_client.controllers.logout import logout
from opthub_client.controllers.select import select
from opthub_client.controllers.show_trial import show_trial
from opthub_client.controllers.show_trials import show_trials
from opthub_client.controllers.submit import submit


@click.group(help="OptHub CLI client.", add_help_option=False)
@click.pass_context
def help(ctx: click.Context) -> None:  # noqa: A001, ARG001
    """Show commands usage."""


help.add_command(show_trials, name="show trials")
help.add_command(show_trial, name="show trial")
help.add_command(select)
help.add_command(submit)
help.add_command(login)
help.add_command(logout)
help.add_command(download)



"""Download trials to a file."""

import json
from pathlib import Path

import click

from opthub_client.context.match_selection import MatchSelectionContext
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError
from opthub_client.errors.user_input_error import UserInputError
from opthub_client.models.trial import fetch_trials

# Number of trials to fetch in one request
SIZE_FETCH_TRIALS = 50


@click.command()
@click.option("-c", "--competition", type=str, help="Competition ID.")
@click.option("-m", "--match", type=str, help="Match ID.")
@click.option(
    "-s",
    "--start",
    type=click.IntRange(min=0),
    default=0,
    help="Start trial number",
)
@click.option(
    "-e",
    "--end",
    type=click.IntRange(min=1),
    default=SIZE_FETCH_TRIALS,
    help="End trial number",
)
@click.option("-desc", "--descending", is_flag=True, help="Show trials in descending order")
@click.option("-suc", "--success", is_flag=True, help="Show only successful trials")
@click.pass_context
def download(
    ctx: click.Context,  # noqa: ARG001
    competition: str | None,
    match: str | None,
    start: int,
    end: int,
    descending: bool,
    success: bool,
) -> None:
    """Download trials to a file."""
    try:
        check_current_version_status()
        match_selection_context = MatchSelectionContext()
        selected_match = match_selection_context.get_match(match, competition)
        output_file = Path(f"trials_{selected_match['alias']}.json")
        total_trials = end - start
        # trial_from is 0 and ascending is True, then increment trial_from by 1 because trial number starts from 1.
        start = start + 1 if start == 0 and not descending else start
        with output_file.open("w") as f:
            all_trials = []
            with click.progressbar(  # type: ignore[var-annotated] # opthub-client/issues/99
                length=total_trials,
                label="Downloading trials",
            ) as bar:
                for index, batch_start in enumerate(range(start, end + 1, SIZE_FETCH_TRIALS)):
                    limit = min(SIZE_FETCH_TRIALS, end - batch_start + 1)
                    trials, is_first, is_last = fetch_trials(
                        selected_match["id"],
                        page=index,
                        page_size=SIZE_FETCH_TRIALS,
                        offset=end if descending else start,
                        limit=limit,
                        is_asc=not descending,
                        display_only_success=success,
                    )
                    all_trials.extend(trials)
                    if (is_first and descending) or (is_last and not descending):
                        bar.update(total_trials)
                        break
                    bar.update(limit)
            json.dump(all_trials, f, indent=4)

        click.echo(f"Trials have been written to {output_file}")
    except (AuthenticationError, FetchError, QueryError, CacheIOError, UserInputError) as error:
        error.handler()
    except Exception:
        click.echo("Unexpected error occurred. Please try again later.")



"""This module contains the functions related to history command."""

import click

from opthub_client.context.match_selection import MatchSelectionContext
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError
from opthub_client.errors.user_input_error import UserInputError
from opthub_client.models.trial import fetch_trial
from opthub_client.view.display_trials import display_trial


@click.command(name="trial")
@click.option("-c", "--competition", type=str, help="Competition ID.")
@click.option("-m", "--match", type=str, help="Match ID.")
@click.option("-d", "--detail", is_flag=True, help="Show detailed information")
@click.argument("trial_no", type=int, required=True)
@click.pass_context
def show_trial(
    ctx: click.Context,  # noqa: ARG001
    competition: str | None,
    match: str | None,
    detail: bool,
    trial_no: int,
) -> None:
    """Check submitted solution."""
    try:
        check_current_version_status()
        match_selection_context = MatchSelectionContext()
        selected_match = match_selection_context.get_match(match, competition)
        # display batch of solutions.
        trial = fetch_trial(selected_match["id"], trial_no=trial_no)
        display_trial(trial, detail)
    except (AuthenticationError, FetchError, QueryError, CacheIOError, UserInputError) as error:
        error.handler()
    except Exception:
        click.echo("Unexpected error occurred. Please try again later.")



"""This module contains the functions related to history command."""

import asyncio

import click
from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent

from opthub_client.context.match_selection import MatchSelectionContext
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError
from opthub_client.errors.user_input_error import UserInputError
from opthub_client.models.trial import fetch_trials_async
from opthub_client.view.display_trials import display_trials, user_interaction_message, user_interaction_message_style


@click.command(name="trials")
@click.option("-c", "--competition", type=str, help="Competition ID.")
@click.option("-m", "--match", type=str, help="Match ID.")
@click.option("-d", "--detail", is_flag=True, help="Show detailed information")
@click.option(
    "-s",
    "--size",
    type=click.IntRange(1, 50),  # show size of trials must be 1-50
    default=20,
    help="Number of trials to display (1-50).",
)
@click.option(
    "-f",
    "--from",
    "trial_from",
    type=int,
    default=0,
    help="Number of trials to display (1-50).",
)
@click.option("-asc", "--ascending", is_flag=True, help="Show trials in ascending order")
@click.option("-suc", "--success", is_flag=True, help="Show only successful trials")
@click.pass_context
def show_trials(
    ctx: click.Context,  # noqa: ARG001
    competition: str | None,
    match: str | None,
    size: int,
    trial_from: int,
    detail: bool,
    ascending: bool,
    success: bool,
) -> None:
    """Check submitted solutions."""
    try:
        check_current_version_status()
        match_selection_context = MatchSelectionContext()
        selected_match = match_selection_context.get_match(match, competition)
        bindings = KeyBindings()
        page = 0
        has_all_trials_displayed = False
        tasks = []
        # trial_from is 0 and ascending is True, then increment trial_from by 1 because trial number starts from 1.
        trial_from = trial_from + 1 if trial_from == 0 and ascending else trial_from

        async def next_trials() -> None:
            nonlocal page, has_all_trials_displayed
            page += 1
            has_more_trials = await fetch_and_display_trials(
                selected_match["id"],
                page,
                size,
                trial_from,
                detail,
                ascending,
                success,
            )
            if not has_more_trials:
                # No more trials, display the message and set the flag.
                has_all_trials_displayed = True

        # n key is to display next batch of solutions
        @bindings.add("n")
        def add_trials(event: KeyPressEvent) -> None:  # noqa: ARG001
            """Display next batch of solutions."""
            if not has_all_trials_displayed:
                task = asyncio.create_task(next_trials())
                tasks.append(task)
            else:
                run_in_terminal(lambda: click.echo("No more trials."), render_cli_done=False)

        @bindings.add("e")  # e for exit
        @bindings.add("q")  # q for exit
        @bindings.add("c-c")  # Ctrl-C for exit
        def exit_trials_view(event: KeyPressEvent) -> None:  # noqa: ARG001
            """Exit the application."""
            session.app.exit()

        # Initialize prompt session with key bindings.
        session: PromptSession[str] = PromptSession(key_bindings=bindings)

        # The initial call to display next batch of solutions async.
        has_more_trials = asyncio.run(
            fetch_and_display_trials(
                selected_match["id"],
                page,
                size,
                trial_from,
                detail,
                ascending,
                success,
            ),
        )
        if not has_more_trials:
            has_all_trials_displayed = True
        # Prompt the user for more solutions(n key) or exit(e or q or Ctrl+c).
        session.prompt(
            HTML(user_interaction_message()),
            key_bindings=bindings,
            style=user_interaction_message_style(),
        )
    except (AuthenticationError, FetchError, QueryError, CacheIOError, UserInputError) as error:
        error.handler()


async def fetch_and_display_trials(
    selected_match_id: str,
    page: int,
    size: int,
    trial_from: int,
    detail: bool,
    asc: bool,
    success: bool,
) -> bool:
    """Fetch and display the trials. Returns True if more trials are available, False if no more trials.

    Args:
        selected_match_id (str): selected match id
        page (int): Page number
        size (int): Number of trials to display
        trial_from (int): Trial number to start from
        detail (bool): True for detailed information, false for general information
        asc (bool): True for show trials in ascending order, false for descending order
        success (bool): True for show only successful trials, false for all trials

    Returns:
        bool: True if more trials are available, False if no more trials
    """
    trials, is_first, is_last = await fetch_trials_async(
        match_id=selected_match_id,
        page=page,
        page_size=size,
        limit=size,
        offset=trial_from,
        is_asc=asc,
        display_only_success=success,
    )
    run_in_terminal(lambda: display_trials(trials, detail), render_cli_done=False)
    if (asc and is_last) or (not asc and is_first):
        run_in_terminal(lambda: click.echo("No more trials."), render_cli_done=False)
        return False
    return True



"""This module contains the FetchError class."""

import sys

import click


class FetchError(Exception):
    """Exception raised for errors in the Fetch execution."""

    def __init__(self, message: str) -> None:
        """Initialize the FetchError class."""
        self.message = message
        super().__init__(self.message)

    def handler(self) -> None:
        """Handle the Fetch error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the solution validator."""

import json


class SolutionValidator:
    """The solution validator."""

    @staticmethod
    def check_solution(raw_solution: str) -> bool:
        """Check validate the solution.

        Args:
            raw_solution (str): The raw data of solution.

        Returns:
            bool: True if the solution is valid, False otherwise.
        """
        try:
            variable = json.loads(raw_solution)
            # Check if variable is a scalar number (int or float)
            if isinstance(variable, (int | float)):
                return True
            # Check if variable is a list of numbers
            if isinstance(variable, list):
                # Each element of variable is float or int
                variable = list(map(float, variable))
                return all(isinstance(value, (float)) for value in variable)
        except (ValueError, json.JSONDecodeError):
            return False
        else:
            # If variable is not a scalar number or a list of numbers return false
            return False



"""This module contains the types and functions related to participant trials."""

from typing import Any, Literal, TypedDict

from gql import gql
from graphql import DocumentNode

from opthub_client.errors.graphql_error import GraphQLError
from opthub_client.errors.query_error import QueryError
from opthub_client.graphql.client import execute_graphql, execute_graphql_async


class Solution(TypedDict):
    """This class represents the solution type."""

    variable: str
    created_at: str


class Evaluation(TypedDict):
    """This class represents the evaluation type."""

    status: Literal["Success", "Failed"]
    objective: float | list[float]
    constraint: float | list[float]
    info: object
    started_at: str
    finished_at: str


class Score(TypedDict):
    """This class represents the score type."""

    status: Literal["Success", "Failed"]
    score: float | None
    started_at: str | None
    finished_at: str | None


class Trial(TypedDict):
    """This class represents the trial type."""

    trialNo: int
    solution: Solution
    status: Literal["evaluating", "success", "scoring", "evaluator_failed", "scorer_failed"]
    evaluation: Evaluation | None
    score: Score | None


def create_trial(trial_data: dict[str, Any]) -> Trial:
    """Create a Trial object from trial data.

    Args:
        trial_data (dict[str, Any]): trail data from the server

    Returns:
        Trial: the trial object
    """
    status = trial_data.get("status", "evaluating")
    solution = Solution(
        variable=trial_data["solution"]["variable"],
        created_at=trial_data["solution"]["createdAt"],
    )
    evaluation = None
    if status in ("success", "scoring", "scorer_failed") and "evaluation" in trial_data:
        evaluation = Evaluation(
            status=trial_data["evaluation"]["status"],
            objective=trial_data["evaluation"]["objective"],
            constraint=trial_data["evaluation"]["constraint"],
            info=trial_data["evaluation"]["info"],
            started_at=trial_data["evaluation"]["startedAt"],
            finished_at=trial_data["evaluation"]["finishedAt"],
        )
    score = None
    if status in ("success") and "score" in trial_data:
        score = Score(
            status=trial_data["score"]["status"],
            score=trial_data["score"]["value"],
            started_at=trial_data["score"]["startedAt"],
            finished_at=trial_data["score"]["finishedAt"],
        )
    return Trial(
        trialNo=trial_data["trialNo"],
        solution=solution,
        status=status,
        evaluation=evaluation,
        score=score,
    )


def fetch_trial(match_id: str, trial_no: int) -> Trial | None:
    """Fetch the history of the user's submitted solution and their evaluation and score.

    Args:
        match_id (str): Match ID in the competition
        trial_no (int): Trial number

    Returns:
        Trial:
            The the history of the user's submitted solution and their evaluation and score.
    """
    query = gql("""
            query getMatchTrialByParticipant(
            $match: MatchIdentifierInput!,
            $trialNo: Int!
            ) {
            getMatchTrialByParticipant(
                match: $match,
                trialNo: $trialNo
            ) {
                trialNo
                status
                solution {
                    variable
                    createdAt
                }
                evaluation {
                    constraint
                    feasible
                    objective
                    status
                    startedAt
                    finishedAt
                    info
                }
                score {
                    status
                    startedAt
                    finishedAt
                    value
                }
            }}""")
    try:
        result = execute_graphql(query, variables={"match": {"id": match_id}, "trialNo": trial_no})
    except GraphQLError as e:
        raise QueryError(resource="trial", detail=str(e.message)) from e
    if result is None:
        return None
    raw_trial = result.get("getMatchTrialByParticipant")
    if not raw_trial:
        return None
    if not isinstance(raw_trial, dict):
        raise QueryError(resource="trial", detail="Invalid data returned.")
    trial = create_trial(raw_trial)
    return trial


def parse_fetched_trials(result: dict[str, Any], display_only_success: bool) -> tuple[list[Trial], bool, bool]:
    """Parse the fetched trials data.

    Args:
        result (Any|None): The fetched trials data
        display_only_success (bool): True to display only successful trials

    Returns:
        tuple[list[Trial], bool, bool]:
            The parsed trials, is_first, and is_last
    """
    data = result.get("getMatchTrialsByParticipant")
    if not isinstance(data, dict):
        raise QueryError(resource="trial", detail="Invalid data returned.")
    is_last = data.get("isLast", False)
    is_first = data.get("isFirst", False)
    trials = []
    for raw_trial in data.get("trials", []):
        trial = create_trial(raw_trial)
        if not display_only_success or trial.get("status") == "success":
            trials.append(trial)
    return trials, is_first, is_last


def make_fetch_trials_query_document() -> DocumentNode:
    """Make the graphql document node to fetch the trials.

    Returns:
        DocumentNode:
            The graphql document node to fetch the trials.
    """
    return gql("""
            query getMatchTrialsByParticipant(
            $match: MatchIdentifierInput!,
            $participant: ParticipantInput,
            $range: MatchTrialsRangeInput,
            $order: Order
            ) {
            getMatchTrialsByParticipant(
                match: $match,
                participant: $participant,
                range: $range,
                order: $order
            ) {
                isFirst
                isLast
                startTrialNo
                endTrialNo
                trials {
                    trialNo
                    status
                    solution {
                        variable
                        createdAt
                    }
                    evaluation {
                        constraint
                        feasible
                        objective
                        status
                        startedAt
                        finishedAt
                        info
                    }
                    score {
                        status
                        startedAt
                        finishedAt
                        value
                    }
                }
            }}""")


def make_fetch_trials_query_variables(
    match_id: str,
    page: int,
    page_size: int,
    limit: int,
    offset: int,
    is_asc: bool,
) -> dict[str, Any]:
    """Make the variables for the fetch trials query.

    Args:
        match_id (str): Match ID in the competition
        page (int): Page number
        page_size(int): Size of the page
        limit (int): Size of the page
        offset (int): Trial number to start fetching
        is_asc (bool): True for show trials in ascending order, False for descending order

    Returns:
         dict[str, Any]:
            The variables for the fetch trials query.
    """
    return {
        "match": {"id": match_id},
        "range": {"startTrialNo": offset + page * page_size, "limit": limit}
        if is_asc
        else {"endTrialNo": offset - page * page_size, "limit": limit},
        "order": "ascending" if is_asc else "descending",
    }


async def fetch_trials_async(
    match_id: str,
    page: int,
    page_size: int,
    limit: int,
    offset: int,
    is_asc: bool,
    display_only_success: bool,
) -> tuple[list[Trial], bool, bool]:
    """Fetch the history of the user's submitted solutions and their evaluations and scores.

    Args:
        match_id (str): Match ID in the competition
        page (int): Page number
        page_size(int): Size of the page
        limit (int): Size of the page
        offset (int): Trial number to start fetching
        is_asc (bool): True for show trials in ascending order, False for descending order
        display_only_success (bool): True to display only successful trials

    Returns:
        list[Trial]:
            The history of the user's submitted solutions and their evaluations and scores.
    """
    query = make_fetch_trials_query_document()
    variables = make_fetch_trials_query_variables(match_id, page, page_size, limit, offset, is_asc)
    try:
        result = await execute_graphql_async(query, variables)
    except GraphQLError as e:
        raise QueryError(resource="trial", detail=str(e.message)) from e
    return parse_fetched_trials(result, display_only_success)


def fetch_trials(
    match_id: str,
    page: int,
    page_size: int,
    limit: int,
    offset: int,
    is_asc: bool,
    display_only_success: bool,
) -> tuple[list[Trial], bool, bool]:
    """Fetch the history of the user's submitted solutions and their evaluations and scores.

    Args:
        match_id (str): Match ID in the competition
        page (int): Page number
        page_size(int): Size of the page
        limit (int): Size of the page
        offset (int): Trial number to start fetching
        is_asc (bool): True for show trials in ascending order, False for descending order
        display_only_success (bool): True to display only successful trials

    Returns:
        list[Trial]:
            The the history of the user's submitted solutions and their evaluations and scores.
    """
    query = make_fetch_trials_query_document()
    variables = make_fetch_trials_query_variables(match_id, page, page_size, limit, offset, is_asc)
    try:
        result = execute_graphql(query, variables)
    except GraphQLError as e:
        raise QueryError(resource="trial", detail=str(e.message)) from e
    return parse_fetched_trials(result, display_only_success)



"""This module contains the functions related to auth command."""

import click

from opthub_client.context.credentials import Credentials
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError


@click.command()
@click.pass_context
def logout(ctx: click.Context) -> None:  # noqa: ARG001
    """Logout."""
    try:
        check_current_version_status()
        credentials = Credentials()
        credentials.clear_credentials()
        click.echo("Successfully logged out.")
    except (CacheIOError, AuthenticationError, QueryError, FetchError) as e:
        e.handler()



"""Read version from toml file."""

import tomllib
from pathlib import Path

from opthub_client.errors.cache_io_error import CacheIOError, CacheIOErrorMessage


def get_version_from_toml() -> str:
    """Read version from pyproject.toml file.

    Returns:
        str: The version of the project as a string.

    Raises:
        VersionNotFoundError: If the version cannot be found or is in an incorrect format.
    """
    path = Path("pyproject.toml")
    with path.open("rb") as file:
        data = tomllib.load(file)
    version = data["project"]["version"]
    if not isinstance(version, str):
        raise CacheIOError(CacheIOErrorMessage.VERSION_FILE_READ_FAILED)
    return version



"""Group of commands to display trials."""

import click

from opthub_client.controllers.show_trial import show_trial
from opthub_client.controllers.show_trials import show_trials


@click.group()
def show() -> None:
    """Group of commands to display trials."""


show.add_command(show_trials)
show.add_command(show_trial)



"""This module contains the class related to match selection context."""

import shelve

from opthub_client.context.utils import get_opthub_client_dir
from opthub_client.errors.cache_io_error import CacheIOError, CacheIOErrorMessage
from opthub_client.errors.user_input_error import UserInputError, UserInputErrorMessage
from opthub_client.models.competition import Competition, fetch_participating_competitions
from opthub_client.models.match import Match, fetch_matches_by_competition

FILE_NAME = "match_selection"


class MatchSelectionContext:
    """The selection context of match."""

    def __init__(self) -> None:
        """Initialize the match selection context with a persistent temporary file."""
        self.file_path = get_opthub_client_dir() / FILE_NAME
        self.load()

    def load(self) -> None:
        """Load the match selection from the shelve file."""
        with shelve.open(str(self.file_path)) as key_store:  # noqa: S301 opthub-client#95
            self.competition_id = key_store.get("competition_id")
            self.match_id = key_store.get("match_id")
            self.match_alias = key_store.get("match_alias")
            self.competition_alias = key_store.get("competition_alias")
            key_store.close()

    def update(self, competition: Competition, match: Match) -> None:
        """Update the match selection in the shelve file.

        Args:
            competition (Competition): Competition instance
            match (Match): Match instance
        """
        with shelve.open(str(self.file_path)) as db:  # noqa: S301 opthub-client#95
            db["competition_id"] = competition["id"]
            db["match_id"] = match["id"]
            db["match_alias"] = match["alias"]
            db["competition_alias"] = competition["alias"]
            db.sync()

    def get_selection(self, match: str | None, competition: str | None) -> tuple[Competition, Match]:
        """Select a competition and match based on the provided aliases.

        This method allows you to select a competition and a match by their aliases.
        If no aliases are provided, it will use default values from `MatchSelectionContext`.
        If the competition or match cannot be found, it raises an appropriate error or exits the program.

        Args:
            match (str | None): The alias of the match to select.
                                If None, it uses the default value from `MatchSelectionContext`.
            competition (str | None): The alias of the competition to select.
                                      If None, it uses the default value from `MatchSelectionContext`.

        Returns:
            tuple[Competition, Match]: A tuple containing the selected competition and match objects.

        """
        match_selection_context = MatchSelectionContext()
        if match is None:
            match = match_selection_context.match_alias
        if competition is None:
            competition = match_selection_context.competition_alias
        if competition is None or match is None:
            raise CacheIOError(CacheIOErrorMessage.MATCH_SELECTION_FILE_READ_FAILED)
        # competitions aliases for choices
        competitions = fetch_participating_competitions()
        selected_competition = next((c for c in competitions if c["alias"] == competition), None)
        if selected_competition is None:
            raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
        matches = fetch_matches_by_competition(selected_competition["id"], selected_competition["alias"])
        selected_match = next((m for m in matches if m["alias"] == match), None)
        if selected_competition is None:
            raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
        if selected_match is None:
            raise UserInputError(UserInputErrorMessage.MATCH_ERROR)
        return selected_competition, selected_match

    def get_match(self, match: str | None, competition: str | None) -> Match:
        """Select a match based on the provided aliases.

        This method allows you to select a match by its alias within a specified competition.
        If no aliases are provided, it will use default values from `MatchSelectionContext`.
        If the competition or match cannot be found, it raises an appropriate error or exits the program.

        Args:
            match (str | None): The alias of the match to select.
                                If None, it uses the default value from `MatchSelectionContext`.
            competition (str | None): The alias of the competition to select.
                                      If None, it uses the default value from `MatchSelectionContext`.

        Returns:
            Match: The selected match object.
        """
        match_selection_context = MatchSelectionContext()
        if match is None:
            match = match_selection_context.match_alias
        if competition is None:
            competition = match_selection_context.competition_alias
        if competition is None or match is None:
            raise CacheIOError(CacheIOErrorMessage.MATCH_NOT_SELECTED)
        competitions = fetch_participating_competitions()
        selected_competition = next((c for c in competitions if c["alias"] == competition), None)
        if selected_competition is None:
            raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
        matches = fetch_matches_by_competition(selected_competition["id"], selected_competition["alias"])
        selected_match = next((m for m in matches if m["alias"] == match), None)
        if selected_competition is None:
            raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
        if selected_match is None:
            raise UserInputError(UserInputErrorMessage.MATCH_ERROR)
        return selected_match



"""This module contains the UserInputError class."""

import sys
from enum import Enum

import click


class UserInputErrorMessage(Enum):
    """Enum class for user input error types."""

    FILE_ERROR = "File input error. Please check the file."
    FILE_PATH_ERROR = "File path error. Please check the file path."
    SOLUTION_ERROR = "Solution input error. Please check the solution."
    MATCH_ERROR = "Match input error. Please check the match."
    COMPETITION_ERROR = "Competition input error. Please check the competition."


class UserInputError(Exception):
    """Exception raised for errors in user input."""

    def __init__(self, error_type: UserInputErrorMessage) -> None:
        """Initialize the UserInputError class."""
        super().__init__(error_type.value)

    def handler(self) -> None:
        """Handle the user input error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the AuthenticationError class."""

import sys
from enum import Enum

import click


class AuthenticationErrorMessage(Enum):
    """Enum class for authentication error types."""

    LOGIN_FAILED = "Login failed. Please check your Username or Password."
    GET_JWKS_PUBLIC_KEY_FAILED = "Failed to get public key. Please try again later."
    LOAD_CREDENTIALS_FAILED = "Not found credentials. Please login with `opt login` command."
    REFRESH_FAILED = "Could not refresh access token. Please login again with `opt login` command."
    DECODE_JWT_TOKEN_FAILED = "Failed to decode JWT token. Please try again later."  # noqa: S105


class AuthenticationError(Exception):
    """Exception raised for authentication related errors."""

    def __init__(self, error_type: AuthenticationErrorMessage) -> None:
        """Initialize the AuthenticationError class."""
        super().__init__(error_type.value)

    def handler(self) -> None:
        """Handle the GraphQL error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the MutationError class."""

import sys
from enum import Enum

import click


class Method(Enum):
    """Enum class for mutation methods."""

    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


class MutationError(Exception):
    """Exception raised for errors during mutations."""

    def __init__(self, method: Method, resource: str, detail: str) -> None:
        """Initialize the MutationError class."""
        self.method = method
        self.resource = resource
        self.detail = detail
        self.message = f"MutationError: {method} {resource} - {detail}"
        super().__init__(self.message)

    def handler(self) -> None:
        """Handle the mutation error."""
        click.echo(str(self))
        sys.exit(1)



"""This module contains the functions related to competitions."""

from typing import TypedDict

from gql import gql

from opthub_client.errors.graphql_error import GraphQLError
from opthub_client.errors.query_error import QueryError
from opthub_client.graphql.client import execute_graphql


class Competition(TypedDict):
    """This class represents the competition type."""

    id: str
    alias: str


def fetch_participating_competitions() -> list[Competition]:
    """Fetch competitions and matches that the user is participating in.

    Args:
         uid (str): user ID
         username (str): user name
    Returns:
         list[Competition]: Competitions and matches that the user is participating in
    Raises:
        ValueError: If no competitions are found for the user or the fetch fails.
    """
    query = gql("""
        query getCompetitionsByParticipantUser(
        $id: String,
        $name: String
        ) {
            getCompetitionsByParticipantUser(
                id: $id,
                name: $name
            ) {
                participating {
                    id
                    alias
                }
            }
        }
        """)
    try:
        result = execute_graphql(query)
    except GraphQLError as e:
        raise QueryError(resource="competitions", detail=str(e)) from e
    data = result.get("getCompetitionsByParticipantUser")
    if not data:
        raise QueryError(resource="competitions", detail="No data returned.")
    participating_competitions = data.get("participating")
    if not isinstance(participating_competitions, list):
        raise QueryError(resource="competitions", detail="Invalid data returned.")
    return [Competition(id=comp["id"], alias=comp["alias"]) for comp in participating_competitions]



"""Util for opthub_client."""

import sys

import click

from opthub_client.context.version import get_version_from_toml
from opthub_client.models.remote_message import get_version_status_messages


def check_current_version_status() -> None:
    """This function gets the version message."""
    version = get_version_from_toml()
    messages = get_version_status_messages(version)
    exit_flag = False
    for message in messages:
        if message.label == "Error":
            click.echo(
                click.style(message.label, fg=message.label_color)
                + ": "
                + click.style(
                    message.message,
                    fg=message.message_color,
                ),
            )
            exit_flag = True
        else:
            click.echo(
                click.style(message.label, fg=message.label_color)
                + ": "
                + click.style(
                    message.message,
                    fg=message.message_color,
                ),
            )
    if exit_flag:
        sys.exit(1)



"""This module contains the functions that display the trials to the user."""

import click
from prompt_toolkit.styles import Style

from opthub_client.models.trial import Trial


def display_trial(trial: Trial | None, is_detail: bool) -> None:
    """Display the trial."""
    if trial is None:
        click.echo("No more solutions to display.")
    else:
        lines = get_trial_info_detail(trial) if is_detail else get_trial_info_general(trial)
        click.echo(lines.rstrip())


def display_trials(trials: list[Trial], is_detail: bool) -> None:
    """Display the trials."""
    if len(trials) == 0:
        click.echo("No solutions to display.")
    else:
        lines = ""
        for trial in trials:
            lines += get_trial_info_detail(trial) if is_detail else get_trial_info_general(trial)
        click.echo(lines.rstrip())


def user_interaction_message() -> str:
    """User interaction message for show trials."""
    return '<n style="class:n">n</n>: next solutions, <e style="class:e">e</e>: exit \n'


def user_interaction_message_style() -> Style:
    """Style for user interaction message."""
    style = Style.from_dict(
        {
            "n": "fg:blue bold",
            "e": "fg:red bold",
        },
    )
    return style


def get_trial_info_general(trial: Trial) -> str:
    """Display the general information of a trial.

    Args:
        trial (Trial): The trial to display.

    Returns:
        str: The general information of the trial.
    """
    lines = (
        f"Trial No: {trial['trialNo']}, status: {trial['status']}, Score: {trial['score']['score']}\n"
        if trial["score"]
        else f"Trial No: {trial['trialNo']}, status: {trial['status']}\n"
    )

    return lines


def get_trial_info_detail(trial: Trial) -> str:
    """Display the general information of a trial.

    Args:
        trial (Trial): The trial to display.

    Returns:
        str: The general information of the trial.
    """
    lines = ""
    lines += "Trial No: " + str(trial["trialNo"]) + "\n"
    lines += "status: " + str(trial["status"]) + "\n"
    lines += "Solution: " + str(trial["solution"]["variable"]) + "\n"
    lines += "Evaluation: " + str(trial["evaluation"]["objective"]) + "\n" if trial["evaluation"] else ""
    lines += "Score: " + str(trial["score"]["score"]) + "\n" if trial["score"] else ""
    return lines



"""This module contains the methods for context."""

from pathlib import Path


def get_opthub_client_dir() -> Path:
    """Get the opthub client directory."""
    opthub_dir = Path.home() / ".opthub_client"
    opthub_dir.mkdir(exist_ok=True)  # Create the directory if it doesn't exist
    return opthub_dir



"""Solution model."""

from gql import gql

from opthub_client.errors.graphql_error import GraphQLError
from opthub_client.errors.mutation_error import Method, MutationError
from opthub_client.graphql.client import execute_graphql


def create_solution(match_id: str, variable: str) -> None:
    """Create a solution by AppSync endpoint.

    Args:
        match_id (str): The match ID.
        variable (object): The variable of solution.
    """
    mutation = gql("""
    mutation createSolution(
        $matchId: String!,
        $variable: AWSJSON!) {
        createSolution(
            matchId: $matchId,
            variable: $variable
        ) {
            matchId
            participantType
            participantId
            trialNo
        }
    }
    """)
    solution_input = {
        "matchId": match_id,
        "variable": variable,
    }
    try:
        execute_graphql(mutation, solution_input)
    except GraphQLError as e:
        raise MutationError(method=Method.CREATE, resource="solution", detail="Failed to submit solutions") from e



"""This module contains the functions related to matches."""

from typing import TypedDict

from gql import gql

from opthub_client.errors.graphql_error import GraphQLError
from opthub_client.errors.query_error import QueryError
from opthub_client.graphql.client import execute_graphql


class Match(TypedDict):
    """This class represents the match type."""

    id: str
    alias: str


def fetch_matches_by_competition(comp_id: str, comp_alias: str) -> list[Match]:
    """Fetch matches by competition alias.

    Args:
        comp_id (str): Competition ID
        comp_alias (str): Competition alias

    Returns:
        list[Match]: Matches related to the competition
    """
    query = gql("""
        query getMatchesByCompetition(
        $id: String,
        $alias: String
        ) {
        getMatchesByCompetition(
            id: $id,
            alias: $alias
        ) {
            id
            competition {
                id
                alias
            }
            problem {
                id
                alias
            }
            indicator {
                id
                alias
            }
            alias
            successTrialsBudget
            submissionsBudget
            isTutorial
            problemPublicEnvironments {
                key
                value
            }
            indicatorPublicEnvironments {
                key
                value
            }
            problemPrivateEnvironments {
                key
                value
            }
            indicatorPrivateEnvironments {
                key
                value
            }
            openAt
            closeAt
        }
        }""")
    try:
        result = execute_graphql(query, variables={"id": comp_id, "alias": comp_alias})
    except GraphQLError as e:
        raise QueryError(resource="matches", detail=str(e.message)) from e
    data = result.get("getMatchesByCompetition")
    if not isinstance(data, list):
        raise QueryError(resource="matches", detail="Invalid data returned.")
    return [Match(id=match["id"], alias=match["alias"]) for match in data]



"""This module contains the functions related to select command."""

import click
from InquirerPy import prompt  # type: ignore[attr-defined]

from opthub_client.context.match_selection import MatchSelectionContext
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError
from opthub_client.errors.user_input_error import UserInputError, UserInputErrorMessage
from opthub_client.models.competition import Competition, fetch_participating_competitions
from opthub_client.models.match import Match, fetch_matches_by_competition

custom_style = {
    "question": "fg:#ffff00 bold",  # question text style
    "answer": "fg:#f44336 bold",  # answer text style
    "input": "fg:#ffeb3b",  # input field style
    "pointer": "fg:#00bcd4 bold",  # select pointer style
    "highlighted": "fg:#00bcd4 bold",  # highlighted item style
    "selected": "fg:#cddc39 bold",  # selected text style
    "instruction": "",  # instruction text style
    "text": "",  # normal text style
}


@click.command()
@click.option("-c", "--competition", type=str, help="Competition ID.")
@click.option("-m", "--match", type=str, help="Match ID.")
def select(
    competition: str | None,
    match: str | None,
) -> None:
    """Select a competition and match.

    Args:
        competition (str | None): option for competition(-c or --competition)
        match (str | None): option for match(-m or --match)
    """
    try:
        check_current_version_status()
        match_selection_context = MatchSelectionContext()
        # competitions aliases for choices
        competitions = fetch_participating_competitions()
        if len(competitions) == 0:
            click.echo("No competitions found that you are participating in.")
            return
        selected_competition = select_competition(competitions, competition)
        matches = fetch_matches_by_competition(selected_competition["id"], selected_competition["alias"])
        selected_match = select_match(matches, match)
        match_selection_context.update(selected_competition, selected_match)
        # show selected competition and match
        click.echo(f"You have selected {selected_competition['alias']}/{selected_match['alias']}")
    except (AuthenticationError, FetchError, QueryError, CacheIOError, UserInputError) as error:
        error.handler()
    except Exception:
        click.echo("Unexpected error occurred. Please try again later.")


def select_competition(competitions: list[Competition], competition_option: str | None) -> Competition:
    """Select a competition.

    Args:
        competitions (list[Competition]): fetch participating competitions
        competition_option (str | None): option for competition(-c or --competition)

    Raises:
        UserInputError: competition error

    Returns:
        Competition: competition
    """
    competition_aliases = [competition["alias"] for competition in competitions]
    # if not set -c commands option
    if competition_option is None:
        competition_questions = [
            {
                "type": "list",
                "message": "Select a competition:",
                "name": "competition",
                "choices": competition_aliases,
            },
        ]
        result_competition = prompt(questions=competition_questions, style=custom_style)
        if isinstance(result_competition["competition"], str):
            competition_option = result_competition["competition"]
        else:
            raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
    if competition_option not in competition_aliases:
        raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
    selected_competition = next((c for c in competitions if c["alias"] == competition_option), None)
    if selected_competition is None:
        raise UserInputError(UserInputErrorMessage.COMPETITION_ERROR)
    return selected_competition


def select_match(matches: list[Match], match_option: str | None) -> Match:
    """Select a match.

    Args:
        matches (list[Match]): fetch matches
        match_option (str | None): option for match(-m or --match)

    Raises:
        UserInputError: match error

    Returns:
        Match: Match
    """
    match_aliases = [match["alias"] for match in matches]
    # if not set -m commands option
    if match_option is None:
        match_questions = [
            {
                "type": "list",
                "message": "Select a match:",
                "name": "match",
                "choices": match_aliases,
            },
        ]
        result_match = prompt(questions=match_questions, style=custom_style)
        if isinstance(result_match["match"], str):
            match = result_match["match"]
        else:
            raise UserInputError(UserInputErrorMessage.MATCH_ERROR)
    if match not in match_aliases:
        raise UserInputError(UserInputErrorMessage.MATCH_ERROR)
    selected_match = next((m for m in matches if m["alias"] == match), None)
    if selected_match is None:
        raise UserInputError(UserInputErrorMessage.MATCH_ERROR)
    return selected_match



"""This module contains the functions related to submit command."""

from pathlib import Path

import click
from InquirerPy import prompt  # type: ignore[attr-defined]
from InquirerPy.validator import PathValidator

from opthub_client.context.match_selection import MatchSelectionContext
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.mutation_error import MutationError
from opthub_client.errors.user_input_error import UserInputError, UserInputErrorMessage
from opthub_client.models.solution import create_solution
from opthub_client.validators.solution import SolutionValidator


@click.command()
@click.option(
    "-c",
    "--competition",
    type=str,
    help="Competition ID",
)
@click.option(
    "-m",
    "--match",
    type=str,
    help="Match ID",
)
@click.option(
    "-f",
    "--file",
    is_flag=True,
    help="Flag to indicate file submission.",
)
@click.pass_context
def submit(ctx: click.Context, match: str | None, competition: str | None, file: bool) -> None:  # noqa: ARG001
    """Submit a solution."""
    try:
        check_current_version_status()
        match_selection_context = MatchSelectionContext()
        selected_competition, selected_match = match_selection_context.get_selection(match, competition)
        raw_solution_value = get_file_submission() if file else get_text_submission()
        if not SolutionValidator.check_solution(raw_solution_value):
            raise UserInputError(UserInputErrorMessage.SOLUTION_ERROR)
        click.echo(
            f"Submitting to {selected_competition['alias']}/{selected_match['alias']}...",
        )
        create_solution(selected_match["id"], raw_solution_value)
        click.echo("...Submitted.")
    except (AuthenticationError, MutationError, CacheIOError, UserInputError) as error:
        error.handler()
    except Exception:
        click.echo("Unexpected error occurred. Please try again later.")


def get_file_submission() -> str:
    """Get solution from the file submission.

    Raises:
        UserInputError: If the solution is not provided.

    Returns:
        str: solution text
    """
    questions = [
        {
            "name": "file",
            "type": "filepath",
            "message": "Submit the solution file (must be a JSON file):",
            "default": str(Path("~/")),
            "validate": PathValidator(is_file=True, message="Input is not a file"),
            "only_files": True,
        },
    ]
    result = prompt(questions)
    if not isinstance(result, dict):
        raise UserInputError(UserInputErrorMessage.FILE_ERROR)
    file_path = result.get("file")
    if not isinstance(file_path, str):
        raise UserInputError(UserInputErrorMessage.FILE_PATH_ERROR)
    full_path = Path(file_path).expanduser()
    return full_path.read_text()


def get_text_submission() -> str:
    """Get solution from the text submission.

    Raises:
        UserInputError: If the solution is not provided.

    Returns:
        str: solution text
    """
    questions = [
        {
            "name": "solution",
            "type": "input",
            "message": "Write the solution:",
        },
    ]
    result = prompt(questions)
    if not isinstance(result, dict) or "solution" not in result:
        raise UserInputError(UserInputErrorMessage.SOLUTION_ERROR)
    return str(result["solution"])



"""Cipher Suite Class for encrypting and decrypting data."""

import shelve

from cryptography.fernet import Fernet

from opthub_client.context.utils import get_opthub_client_dir

FILE_NAME = "encryption_key"


class CipherSuite:
    """Cipher suite class for encrypt."""

    def __init__(self) -> None:
        """Initialize the credentials context with a persistent temporary file."""
        self.file_path = get_opthub_client_dir() / FILE_NAME

    def get(self) -> Fernet:
        """Get the Fernet cipher suite using the encryption key.

        Returns:
            Fernet: Fernet cipher suite
        """
        encryption_key = self.load_or_generate_key()
        return Fernet(encryption_key)

    def load_or_generate_key(self) -> bytes:
        """Load the encryption key from the shelve file, or generate a new one if it doesn't exist.

        Returns:
            bytes: encryption key
        """
        with shelve.open(str(self.file_path)) as key_store:  # noqa: S301 opthub-client#95
            key = key_store.get("encryption_key")
            if key is None:
                key = Fernet.generate_key()
                key_store["encryption_key"] = key
                key_store.sync()
            return key

    def encrypt(self, data: str) -> bytes:
        """Encrypt the data.

        Args:
            data (str): data to encrypt

        Returns:
            bytes: encrypted data
        """
        cipher_suite = self.get()
        return cipher_suite.encrypt(data.encode())

    def decrypt(self, data: bytes) -> str:
        """Decrypt the data.

        Args:
            data (bytes): data to decrypt

        Returns:
            str: decrypted data
        """
        cipher_suite = self.get()
        return cipher_suite.decrypt(data).decode()



"""This module contains the functions related to auth command."""

import click

from opthub_client.context.credentials import Credentials
from opthub_client.controllers.utils import check_current_version_status
from opthub_client.errors.authentication_error import AuthenticationError
from opthub_client.errors.cache_io_error import CacheIOError
from opthub_client.errors.fetch_error import FetchError
from opthub_client.errors.query_error import QueryError


@click.command()
@click.option("--username", "-u", "username", required=True, prompt=True)
@click.option("--password", "-p", "password", prompt=True, hide_input=True)
@click.pass_context
def login(ctx: click.Context, username: str, password: str) -> None:  # noqa: ARG001
    """Login."""
    try:
        credentials = Credentials()
        credentials.cognito_login(username, password)
        click.echo(f"Hello {username}. Successfully logged in.")
        check_current_version_status()
    except (CacheIOError, AuthenticationError, QueryError, FetchError) as e:
        e.handler()



"""This module contains the GraphQL client getter."""

from typing import Any

from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError
from graphql import DocumentNode

from opthub_client.context.credentials import Credentials
from opthub_client.errors.graphql_error import GraphQLError

URL = "https://jciqso7l7rhajfkt5s3dhybpcu.appsync-api.ap-northeast-1.amazonaws.com/graphql"


def get_gql_client() -> Client:
    """Get the GraphQL client.

    Returns:
        Client: The GraphQL client

    Raises:
        AuthenticationError: If authentication fails
    """
    credentials = Credentials()
    credentials.load()
    headers = {"Authorization": f"Bearer {credentials.access_token}"}
    transport = AIOHTTPTransport(url=URL, headers=headers)
    return Client(transport=transport, fetch_schema_from_transport=True)


def execute_graphql(request: DocumentNode, variables: dict[str, Any] | None = None) -> dict[str, Any]:
    """Execute a graphql request.

    Args:
        request (DocumentNode): graphql client request
        variables (_type_, optional): query variables. Defaults to None.

    Raises:
        GraphQLError: graphql error

    Returns:
        dict[str, Any]: result
    """
    client = get_gql_client()
    try:
        return client.execute(request, variable_values=variables)
    except TransportQueryError as auth_error:
        error_message = auth_error.errors[0]["message"] if auth_error.errors else "Unexpected error"
        raise GraphQLError(message=error_message) from auth_error


async def execute_graphql_async(request: DocumentNode, variables: dict[str, Any] | None = None) -> dict[str, Any]:
    """Execute a graphql request.

    Args:
        request (DocumentNode): graphql client request
        variables (_type_, optional): query variables. Defaults to None.

    Raises:
        GraphQLError: graphql error

    Returns:
        dict[str, Any]: result
    """
    client = get_gql_client()
    try:
        return await client.execute_async(request, variable_values=variables)
    except TransportQueryError as auth_error:
        error_message = auth_error.errors[0]["message"] if auth_error.errors else "Unexpected error"
        raise GraphQLError(message=error_message) from auth_error



"""Fetching messages for display in OptHub Client."""

from gql import gql

from opthub_client.errors.graphql_error import GraphQLError
from opthub_client.errors.query_error import QueryError
from opthub_client.graphql.client import execute_graphql


class RemoteMessage:
    """Remote Message."""

    label: str
    label_color: str
    message: str
    message_color: str

    def __init__(self, label: str, label_color: str, message: str, message_color: str) -> None:
        """Initialize the RemoteMessage class.

        Args:
            label (str): label
            label_color (str): color of the label
            message (str): message
            message_color (str): color of the message
        """
        self.label = label
        self.label_color = label_color
        self.message = message
        self.message_color = message_color


def get_version_status_messages(version: str) -> list[RemoteMessage]:
    """Get messages for display in OptHub Client.

    Returns:
        dict[str, str]: The messages.
    """
    query = gql("""
    query getCLIVersionStatus($version: String) {
    getCLIVersionStatus(version: $version) {
        label
        labelColor
        message
        messageColor
    }
    }
    """)
    try:
        result = execute_graphql(query, variables={"version": version})
    except GraphQLError as e:
        raise QueryError(resource="version status", detail=str(e.message)) from e
    data = result.get("getCLIVersionStatus")
    if not data:
        raise QueryError(resource="version status", detail="No data returned.")
    if not isinstance(data, list):
        raise QueryError(resource="version status", detail="Invalid data returned.")
    return [
        RemoteMessage(
            label=item["label"],
            label_color=item["labelColor"],
            message=item["message"],
            message_color=item["messageColor"],
        )
        for item in data
    ]



"""This module contains the class related to credentials."""

import shelve
import time
from pathlib import Path
from typing import Any

import boto3
import jwt
import requests
from jwcrypto import jwk  # type: ignore[import-untyped]

from opthub_client.context.cipher_suite import CipherSuite
from opthub_client.context.utils import get_opthub_client_dir
from opthub_client.errors.authentication_error import AuthenticationError, AuthenticationErrorMessage

FILE_NAME = "credentials"
CLIENT_ID = "24nvfsrgbuvu75h4o8oj2c2oek"
JWKS_URL = "https://cognito-idp.ap-northeast-1.amazonaws.com/ap-northeast-1_1Y69fktA0/.well-known/jwks.json"


class Credentials:
    """The credentials class. To store and manage the credentials."""

    file_path: Path
    access_token: str | None
    refresh_token: str | None
    expire_at: str | None
    uid: str | None
    username: str | None

    def __init__(self) -> None:
        """Initialize the credentials context with a persistent temporary file."""
        self.file_path = get_opthub_client_dir() / FILE_NAME

    def load(self) -> None:
        """Load the credentials from the shelve file."""
        try:
            cipher_suite = CipherSuite()
            with shelve.open(str(self.file_path)) as key_store:  # noqa: S301 opthub-client#95
                # decrypt the credentials
                self.access_token = cipher_suite.decrypt(key_store.get("access_token", b""))
                self.refresh_token = cipher_suite.decrypt(key_store.get("refresh_token", b""))
                self.expire_at = cipher_suite.decrypt(key_store.get("expire_at", b""))
                self.uid = cipher_suite.decrypt(key_store.get("uid", b""))
                self.username = cipher_suite.decrypt(key_store.get("username", b""))
                key_store.close()
        except Exception as e:
            self.clear_credentials()
            raise AuthenticationError(AuthenticationErrorMessage.LOAD_CREDENTIALS_FAILED) from e
        if self.is_expired():
            self.refresh_access_token()

    def update(self, access_token: str, refresh_token: str) -> None:
        """Update the credentials in the shelve file."""
        cipher_suite = CipherSuite()
        with shelve.open(str(self.file_path)) as key_store:  # noqa: S301 opthub-client#95
            # encrypt the credentials
            key_store["access_token"] = cipher_suite.encrypt(access_token)
            key_store["refresh_token"] = cipher_suite.encrypt(refresh_token)
            # decode the access token to get the expire time, user id and user name
            public_key = self.get_jwks_public_key(access_token)
            token = self.decode_jwt_token(access_token, public_key)
            key_store["expire_at"] = cipher_suite.encrypt(str(token.get("exp")))
            key_store["uid"] = cipher_suite.encrypt(token.get("sub"))
            key_store["username"] = cipher_suite.encrypt(token.get("username"))
            key_store.sync()
        self.access_token = access_token
        self.refresh_token = refresh_token

    def is_expired(self) -> bool:
        """Determine if the access token has expired.

        Returns:
            bool: True if the token has expired, otherwise False.
        """
        if self.expire_at is None:
            return True
        current_time = int(time.time())
        expire_at_timestamp = int(self.expire_at)
        return current_time > expire_at_timestamp

    def refresh_access_token(self) -> None:
        """Refresh the access token using refresh token.

        Returns:
            bool: true if the access token is refreshed successfully, otherwise false.
        """
        if self.refresh_token is None:
            self.clear_credentials()
            raise AuthenticationError(AuthenticationErrorMessage.REFRESH_FAILED)
        try:
            client = boto3.client("cognito-idp", region_name="ap-northeast-1")
            response = client.initiate_auth(
                AuthFlow="REFRESH_TOKEN_AUTH",
                AuthParameters={"REFRESH_TOKEN": self.refresh_token},
                ClientId=CLIENT_ID,
            )
            access_token = response["AuthenticationResult"]["AccessToken"]
        except Exception as e:
            self.clear_credentials()
            raise AuthenticationError(AuthenticationErrorMessage.REFRESH_FAILED) from e
        self.update(access_token, self.refresh_token)

    def cognito_login(self, username: str, password: str) -> None:
        """Login to cognito user pool. And update the credentials.

        Args:
            username (str): username
            password (str): password
        """
        try:
            client = boto3.client("cognito-idp", region_name="ap-northeast-1")
            response = client.initiate_auth(
                AuthFlow="USER_PASSWORD_AUTH",
                AuthParameters={"USERNAME": username, "PASSWORD": password},
                ClientId=CLIENT_ID,
            )
            access_token = response["AuthenticationResult"]["AccessToken"]
            refresh_token = response["AuthenticationResult"]["RefreshToken"]
        except Exception as e:
            self.clear_credentials()
            raise AuthenticationError(AuthenticationErrorMessage.LOGIN_FAILED) from e
        self.update(access_token, refresh_token)

    def clear_credentials(self) -> None:
        """Clear the credentials in the shelve file."""
        with shelve.open(str(self.file_path)) as key_store:  # noqa: S301 opthub-client#95
            key_store.clear()
            key_store.sync()
        self.access_token = None
        self.refresh_token = None
        self.expire_at = None
        self.uid = None
        self.username = None

    def get_jwks_public_key(self, access_token: str) -> Any:  # noqa: ANN401
        """Get the public key from the JWKS URL.

        Args:
            access_token (str): access token

        Raises:
            ValueError: fail to get JWKS
            ValueError: fail to get public key

        Returns:
            Any: public key
        """
        try:
            response = requests.get(JWKS_URL, timeout=10)  # set timeout 10 seconds
            response.raise_for_status()
        except requests.RequestException as e:
            raise AuthenticationError(AuthenticationErrorMessage.GET_JWKS_PUBLIC_KEY_FAILED) from e
        jwks = response.json()
        headers = jwt.get_unverified_header(access_token)
        kid = headers["kid"]
        for key in jwks["keys"]:
            if key["kid"] == kid:
                jwk_obj = jwk.JWK(**key)
                public_key_pem = jwk_obj.export_to_pem()
                return public_key_pem
        raise AuthenticationError(AuthenticationErrorMessage.GET_JWKS_PUBLIC_KEY_FAILED)

    def decode_jwt_token(self, access_token: str, public_key: bytes) -> Any:  # noqa: ANN401
        """Decode the JWT token.

        Args:
            access_token (str): access token
            public_key (bytes | Any): public key

        Raises:
            ValueError: fail to get JWKS
            ValueError: fail to get public key

        Returns:
            Any: decoded token
        """
        try:
            return jwt.decode(
                access_token,
                public_key,
                algorithms=["RS256"],
                options={"verify_signature": True},
            )
        except Exception as e:
            raise AuthenticationError(AuthenticationErrorMessage.DECODE_JWT_TOKEN_FAILED) from e



"""This module contains the OptHub CLI client entrypoint."""

import click

from opthub_client.controllers.download import download
from opthub_client.controllers.help import help
from opthub_client.controllers.login import login
from opthub_client.controllers.logout import logout
from opthub_client.controllers.select import select
from opthub_client.controllers.show import show
from opthub_client.controllers.submit import submit

custom_style = {
    "question": "fg:#ffff00 bold",  # question text style
    "answer": "fg:#f44336 bold",  # answer text style
    "input": "fg:#ffeb3b",  # input field style
    "pointer": "fg:#00bcd4 bold",  # select pointer style
    "highlighted": "fg:#00bcd4 bold",  # highlighted item style
    "selected": "fg:#cddc39 bold",  # selected text style
    "instruction": "",  # instruction text style
    "text": "",  # normal text style
}


@click.group(help="OptHub CLI client.")
@click.version_option()
@click.pass_context
def opt(ctx: click.Context) -> None:  # noqa: ARG001
    """This function is for OptHub CLI client entrypoint."""


opt.add_command(show)
opt.add_command(select)
opt.add_command(submit)
opt.add_command(help)
opt.add_command(login)
opt.add_command(logout)
opt.add_command(download)
